
/*################################################################################################
 *##################                                                            ##################
 *##################            DoExec - OKit programs evaluation               ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * GOKe: Interpreter for the OKit language, based on the libOKeval
 * builtins implementation, and the Gtk GUI toolkit.
 *
 * Copyright (C) 2001 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  varkhan@free.fr
 * Homepage: http://varkhan.free.fr/
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <OKit/ErrLog.h>
#include <OKit/ObjTypes.h>
#include <OKit/SysTypes.h>
#include <OKit/FileIOMod.h>
#include <OKit/ParseMod.h>

#include <OKeval/InstSet.h>
#include <DoExec.h>




//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////   File loading and execution
/////

 
// This is a HACK, the blank func was implemented in the OKit Parse Module...
short isblank(char chr) ;


long FileLoad(char * file, unsigned long verb, OKit_Stack stk, OKit_NameSpace dom) {
	unsigned long off = 0 ;
	int fd ;
	int eof, eol ;
	unsigned long ln ;       // line number
	unsigned long buffallc ; // allocated size of buffer
	unsigned long buffused ; // really used size
	unsigned long buffline ; // line length
	unsigned long buffrd ;   // read amount
	char * buffer ;
	char * tbuffer ;

	OKit_List exe = NULL ;
	
	if( (file[0]=='-') && (file[1]==0) ) fd = dup(0) ;
	else fd = open(file,O_RDONLY) ;
	if(fd==-1) {
		fprintf(stderr,"Open failed on %s; aborting\n",file) ;
		return -1 ;
	}
	eof = 0 ;
	ln = 0 ;
	buffallc = 1 ;
	buffused = 0 ;
	buffline = 0 ;
	buffer = malloc(buffallc+1) ;
	if(buffer==NULL) {
		fprintf(stderr,"Failed allocation; aborting\n")	;
		return -1 ;
	}
	while( (!eof) || (buffused>0) ) {
		eol = 0 ;
		// line read into buffer
		while(!eol) {
			buffrd = read(fd,buffer+buffused,buffallc-buffused) ;
			if(buffrd==-1) {
				fprintf(stderr,"Read error in %s, line %lu; aborting\n",file,ln) ;
				return -1 ;
			}
			buffused += buffrd ;
			if(buffused<buffallc) {
				eof = 1 ;
				eol = 1 ;
				if( (buffused==0) || (buffer[buffused-1]!='\n') ) {
					buffer[buffused]='\n' ;
					buffused ++ ;
				}
			}
			else {
				tbuffer = realloc(buffer,(buffallc<<1)+1) ;
				if(tbuffer==NULL) {
					fprintf(stderr,"Failed allocation; aborting\n") ;
					return -1 ;
				}
				buffer = tbuffer ;
				buffallc = (buffallc<<1) ;
			}
		}
		buffline = 0 ;
		while(buffline<buffused) {
			if(buffer[buffline]=='\n') {
				ln ++ ;
				// empty line ?
				if(buffline==0) {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
				// treat escape sequence, one less char
				else if(buffer[buffline-1]=='\\') {
					off = buffline ;
					while(off<buffused) {
						buffer[off-1] = buffer[off] ;
						off ++ ;
					}
					buffused -- ;
				}
				// real end of line
				else {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
			}
			buffline ++ ;
		}
		// line parsing
		buffer[buffline-1] = 0 ;
		if(buffer[0]=='#') {
			if( (buffer[1]=='#') && (verb&PRINT_COM) ) {
				fprintf(stderr,">>%s\n",buffer+2) ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		off = 0 ;
		while(isblank(buffer[off])) {
			if(buffer[off]==0) break ;
			off ++ ;
		}
		if(buffer[off]==0) {
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		errno = 0 ;
		exe = OKit_ListParse(BltTable,BltTable_Size,buffer,&off) ;
//		OKit_ListDup(exe) ;
		// line evaluation
		if( (exe==NULL) && (errno==EINVAL) ) {
			fprintf(stderr,"Syntax error in %s, line %lu: %50.50s...\n",file,ln,buffer) ;
			fprintf(stderr,"...%20.20s%-20.20s...\n",buffer+((off<20)?0:(off-20)),buffer+off) ;
			fprintf(stderr,"                       ^\n") ;
			if(verb&ABORT_ON_ERR) {
				free(buffer) ;
				return -ln ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffer[off] = 0 ;
			buffused -= buffline ;
			continue ;
		}
//		OKit_ListDup(exe) ;
		if(verb&REPORT_ALL) {
			fprintf(stderr,"Loaded ") ;
			fflush(stderr) ;
			OKit_ListPrint(exe,2) ;
			fprintf(stderr," ;\n") ;
		}
		while(OKit_ListLen(exe)>0) {
			OKit_StackPush(stk,OKit_ListShift(exe)) ;
		}
		OKit_ListFree(exe) ;
		if(verb&PRINT_ALL) {
			if(OKit_StackDepth(stk)>0) OKit_ObjectPrint(OKit_StackGet(stk,1),1) ;
			printf("\n") ;
		}
		off = 0 ;
		while(off<buffused-buffline) {
			buffer[off] = buffer[off+buffline] ;
			off ++ ;
		}
		buffused -= buffline ;
	}
	free(buffer) ;
	close(fd) ;
	return ln ;
}


long FileExec(char * file, unsigned long verb, OKit_Stack stk, OKit_NameSpace dom) {
	unsigned long off = 0 ;
	int fd ;
	int eof, eol ;
	unsigned long ln ;       // line number
	unsigned long buffallc ; // allocated size of buffer
	unsigned long buffused ; // really used size
	unsigned long buffline ; // line length
	unsigned long buffrd ;   // read amount
	char * buffer ;
	char * tbuffer ;

	OKit_Status ret ;
	OKit_List exe = NULL ;

	
	if( (file[0]=='-') && (file[1]==0) ) fd = dup(0) ;
	else fd = open(file,O_RDONLY) ;
	if(fd==-1) {
		fprintf(stderr,"Open failed on %s; aborting\n",file) ;
		return -1 ;
	}
	eof = 0 ;
	ln = 0 ;
	buffallc = 1 ;
	buffused = 0 ;
	buffline = 0 ;
	buffer = malloc(buffallc+1) ;
	if(buffer==NULL) {
		fprintf(stderr,"Failed allocation; aborting\n")	;
		return -1 ;
	}
	while( (!eof) || (buffused>0) ) {
		eol = 0 ;
		// line read into buffer
		while(!eol) {
			buffrd = read(fd,buffer+buffused,buffallc-buffused) ;
			if(buffrd==-1) {
				fprintf(stderr,"Read error in %s, line %lu; aborting\n",file,ln) ;
				return -1 ;
			}
			buffused += buffrd ;
			if(buffused<buffallc) {
				eof = 1 ;
				eol = 1 ;
				if( (buffused==0) || (buffer[buffused-1]!='\n') ) {
					buffer[buffused]='\n' ;
					buffused ++ ;
				}
			}
			else {
				tbuffer = realloc(buffer,(buffallc<<1)+1) ;
				if(tbuffer==NULL) {
					fprintf(stderr,"Failed allocation; aborting\n") ;
					return -1 ;
				}
				buffer = tbuffer ;
				buffallc = (buffallc<<1) ;
			}
		}
		buffline = 0 ;
		while(buffline<buffused) {
			if(buffer[buffline]=='\n') {
				ln ++ ;
				// empty line ?
				if(buffline==0) {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
				// treat escape sequence, one less char
				else if(buffer[buffline-1]=='\\') {
					off = buffline ;
					while(off<buffused) {
						buffer[off-1] = buffer[off] ;
						off ++ ;
					}
					buffused -- ;
				}
				// real end of line
				else {
					eol = 1 ;
					buffline ++ ;
					break ;
				}
			}
			buffline ++ ;
		}
		// line parsing
		buffer[buffline-1] = 0 ;
		if(buffer[0]=='#') {
			if( (buffer[1]=='#') && (verb&PRINT_COM) ) {
				fprintf(stderr,">>%s\n",buffer+2) ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		off = 0 ;
		while(isblank(buffer[off])) {
			if(buffer[off]==0) break ;
			off ++ ;
		}
		if(buffer[off]==0) {
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffused -= buffline ;
			continue ;
		}
		errno = 0 ;
		exe = OKit_ListParse(BltTable,BltTable_Size,buffer,&off) ;
		OKit_ListDup(exe) ;
		// line evaluation
		if( (exe==NULL) && (errno==EINVAL) ) {
			fprintf(stderr,"Syntax error in %s, line %lu: %50.50s...\n",file,ln,buffer) ;
			fprintf(stderr,"...%20.20s%-20.20s...\n",buffer+((off<20)?0:(off-20)),buffer+off) ;
			fprintf(stderr,"                       ^\n") ;
			if(verb&ABORT_ON_ERR) {
				free(buffer) ;
				return -ln ;
			}
			off = 0 ;
			while(off<buffused-buffline) {
				buffer[off] = buffer[off+buffline] ;
				off ++ ;
			}
			buffer[off] = 0 ;
			buffused -= buffline ;
			continue ;
		}
//		OKit_ListDup(exe) ;
		errno = 0 ;
		ret = OKit_ListEval(stk,dom,exe) ;
		if(verb&REPORT_ALL) {
			fprintf(stderr,"Evaluated ") ;
			fflush(stderr) ;
			OKit_ListPrint(exe,2) ;
			if(ret.num) fprintf(stderr," ; error %lu at %s > %s\n",ret.num,ret.com,ret.msg) ;
			else fprintf(stderr," ; OK\n") ;
		}
		if(verb&PRINT_ALL) {
			if(OKit_StackDepth(stk)>0) OKit_ObjectPrint(OKit_StackGet(stk,1),1) ;
			printf("\n") ;
		}
		OKit_ListDrop(exe) ;
		OKit_ListFree(exe) ;
		off = 0 ;
		while(off<buffused-buffline) {
			buffer[off] = buffer[off+buffline] ;
			off ++ ;
		}
		buffused -= buffline ;
	}
	free(buffer) ;
	close(fd) ;
	return ln ;
}



long StrExec(char * str, unsigned long verb, OKit_Stack stk, OKit_NameSpace dom) {
	unsigned long off = 0 ;
	OKit_Status ret ;
	OKit_List exe = NULL ;
	errno = 0 ;
	while(isblank(str[off])) {
		if(str[off]==0) break ;
		off ++ ;
	}
	if(str[off]==0) {
		return 0 ;
	}
	exe = OKit_ListParse(BltTable,BltTable_Size,str,&off) ;
	if( (exe==NULL) && (errno==EINVAL) ) {
		fprintf(stderr,"Parsed '%s' ; syntax error\n",str) ;
		return -off ;
	}
	OKit_ListDup(exe) ;
	ret = OKit_ListEval(stk,dom,exe) ;
	if(verb&REPORT_ALL) {
		if(ret.num) {
			fprintf(stderr,"Evaluated '") ;
			fflush(stderr) ;
			OKit_ListPrint(exe,2) ;
			fprintf(stderr,"' ; error %lu at %s > %s\n",ret.num,ret.com,ret.msg) ;
			return -off ;
		}
	}
	if(verb&PRINT_ALL) {
		if(OKit_StackDepth(stk)>0) OKit_ObjectPrint(OKit_StackGet(stk,1),1) ;
		printf("\n") ;
	}
	OKit_ListDrop(exe) ;
	OKit_ListFree(exe) ;
	return off ;
}

