

/*################################################################################################
 *##################                                                            ##################
 *##################            Callbacks - Gtk signals and actions             ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */


/*
 * GOKe: Interpreter for the OKit language, based on the libOKeval
 * builtins implementation, and the Gtk GUI toolkit.
 *
 * Copyright (C) 2001 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  varkhan@free.fr
 * Homepage: http://varkhan.free.fr/
 */


#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <OKit/ErrLog.h>
#include <OKit/SysTypes.h>
#include <OKit/ObjTypes.h>
#include <OKit/FileIOMod.h>


#include <OKeval/InstSet.h>
#include <DoExec.h>



extern GtkWidget *dialog_options;



//################################################################################
//###
//###  Dialogs callbacks
//###


gboolean on_dialog_help_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	if( (event->keyval=='q')&&(event->state&0x4) ) gtk_widget_destroy(user_data) ;
	return FALSE;
}

gboolean on_dialog_about_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	if( (event->keyval=='q')&&(event->state&0x4) ) gtk_widget_destroy(user_data) ;
	return FALSE;
}

gboolean on_dialog_syntax_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	if( (event->keyval=='q')&&(event->state&0x4) ) gtk_widget_destroy(user_data) ;
	return FALSE;
}

void on_clist_syntax_realize (GtkWidget *widget, gpointer user_data) {

	gchar *line0[5] = { "Object",	"",	"Generic object",		"[none]",	"" } ;
	gchar *line1[5] = { "",	"Void",	"Undefined object",		"[null]",	"" } ;
	gchar *line2[5] = { "",	"Intg",	"Integer number on 32 bits",	"[scal]",	"123456" } ;
	gchar *line3[5] = { "",	"Fltp",	"Floating point number",	"[scal]",	"3.1415926e+00" } ;
	gchar *line4[5] = { "",	"Strg",	"Character string",		"[scal]",	"\"any text\"" } ;
	gchar *line5[5] = { "",	"List",	"All-purposes list",		"[list]",	"[ obj1 obj2 ... ]" } ;
	gchar *line6[5] = { "",	"Code",	"Dynamic executable code",	"[exec]",	"<CODE_NAME> or <CODE_NAME:data>" } ;
	gchar *line7[5] = { "",	"Bltn",	"Static executable code",	"[exec]",	"BLTN_NAME" } ;
	gtk_clist_append (GTK_CLIST (widget), line0); 
	gtk_clist_append (GTK_CLIST (widget), line1); 
	gtk_clist_append (GTK_CLIST (widget), line2); 
	gtk_clist_append (GTK_CLIST (widget), line3); 
	gtk_clist_append (GTK_CLIST (widget), line4); 
	gtk_clist_append (GTK_CLIST (widget), line5); 
	gtk_clist_append (GTK_CLIST (widget), line6); 
	gtk_clist_append (GTK_CLIST (widget), line7); 
}

gboolean on_dialog_builtins_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	if( (event->keyval=='q')&&(event->state&0x4) ) gtk_widget_destroy(user_data) ;
	return FALSE;
}

void on_clist_builtins_realize (GtkWidget *widget, gpointer user_data) {
	char * line[2] ;
        unsigned long i = 0 ;
	while(i<BltTable_Size) {
		line[0] = BltTable[i]->name ;
		line[1] = BltTable[i]->desc ;
		gtk_clist_append (GTK_CLIST (widget), line) ;
		i ++ ;
	}
}


//################################################################################
//###
//###  File selector asynchrone handlers
//###


void ( *file_do_func) (gchar *,gpointer) = NULL;
gpointer file_do_data = NULL;

void do_file_browse(gchar *file, void ( *func) (gchar *,gpointer), gpointer data) {
	GtkWidget *file_browse;
	file_do_func = func;
	file_do_data = data;
	file_browse = create_file_browse ();
	gtk_file_selection_set_filename (GTK_FILE_SELECTION (file_browse), file) ;
	gtk_widget_show (file_browse);
}

void on_file_ok_button_clicked (GtkButton *button, gpointer user_data) {
	if(file_do_func!=NULL) {
		file_do_func(
			gtk_file_selection_get_filename(GTK_FILE_SELECTION(user_data)),
			file_do_data) ;
	}
	gtk_widget_destroy (GTK_WIDGET (user_data)) ;
	file_do_func = NULL;
	file_do_data = NULL;
}

void on_file_cancel_button_clicked (GtkButton *button, gpointer user_data) {
	gtk_widget_destroy (GTK_WIDGET (user_data)) ;
	file_do_func = NULL;
	file_do_data = NULL;
}



//################################################################################
//###
//###  Options dialog
//###


struct GOKe_Options {
	unsigned char dbg_levl ;
	unsigned long rep_verb ;
	unsigned long rep_exec ;
} ;

struct GOKe_Options opt_cur = { 0, REPORT_ALL, 0 } ;
struct GOKe_Options opt_edt = { 0, REPORT_ALL, 0 } ;

void on_options_apply(void) {
	opt_cur = opt_edt ;
	fprintf(stderr,"Options:\n") ;
	if(opt_cur.dbg_levl>0) {
		fprintf(stderr,"Debug (%d) - ",opt_cur.dbg_levl) ;
		OKit_ERRTRACEDBG((1<<(opt_cur.dbg_levl-1))-1) ;
	}
	else {
		fprintf(stderr,"Debug off - ") ;
		OKit_ERRTRACEDBG(0) ;
	}
	if(opt_cur.rep_verb&REPORT_ALL) {
		fprintf(stderr," Warn about errors  - ") ;
	}
	else {
		fprintf(stderr,"Don't report errors - ") ;
	}
	if(opt_cur.rep_exec&REPORT_ALL) {
		fprintf(stderr,"  File loading reports  - ") ;
	}
	else {
		fprintf(stderr,"No file loading reports - ") ;
	}
	if(opt_cur.rep_exec&PRINT_COM) {
		fprintf(stderr,"Print comments in files\n") ;
	}
	else {
		fprintf(stderr,"Don't print comments\n") ;
	}
}

void on_button_opt_ok_clicked (GtkButton *button, gpointer user_data) {
	gtk_widget_hide(GTK_WIDGET(user_data));
	on_options_apply() ;
}

void on_button_opt_ap_clicked (GtkButton *button, gpointer user_data) {
	on_options_apply() ;
}

gboolean on_dialog_options_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	if( (event->keyval=='q')&&(event->state&0x4) ) gtk_widget_hide(widget) ;
	else if(event->keyval==GDK_Return) {
		printf("Saving options...\n") ;
		opt_cur = opt_edt ;
		gtk_widget_hide(widget) ;
	}
	return FALSE;
}

void on_checkbutton_dbg_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	if(togglebutton->active) {
		opt_edt.dbg_levl = 1 ;
	}
	else {
		GtkAdjustment *adj ;
	       	adj = gtk_range_get_adjustment (GTK_RANGE (user_data)) ;
		gtk_adjustment_set_value(adj,adj->lower) ;
		opt_edt.dbg_levl = 0 ;
	}
}

gboolean on_hscale_dbg_change (GtkWidget *widget, GdkEventFocus *event, gpointer user_data) {
	if(opt_edt.dbg_levl>0) {
		opt_edt.dbg_levl = gtk_range_get_adjustment (GTK_RANGE (widget))->value+1 ;
	}
	else {
		GtkAdjustment *adj ;
	       	adj = gtk_range_get_adjustment (GTK_RANGE (user_data)) ;
		gtk_adjustment_set_value(adj,adj->lower) ;
	}
	return FALSE;
}


// Show executed
void on_checkbutton_opt_file1_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	if(togglebutton->active) {
		opt_edt.rep_exec |= REPORT_ALL ;
	}
	else {
		opt_edt.rep_exec &= ~REPORT_ALL ;
	}
}

// Warn about syntax
void on_checkbutton_opt_file2_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	if(togglebutton->active) {
		opt_edt.rep_verb |= REPORT_ALL ;
	}
	else {
		opt_edt.rep_verb &= ~REPORT_ALL ;
	}
}

// Display comments
void on_checkbutton_opt_file3_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	if(togglebutton->active) {
		opt_edt.rep_exec |= PRINT_COM ;
	}
	else {
		opt_edt.rep_exec &= ~PRINT_COM ;
	}
}

//################################################################################
//###
//###  OKit_Stack and domain lists, command and console text boxes
//###


OKit_Stack okit_stk;
OKit_NameSpace okit_dom;

GtkWidget *clist_stk;
GtkWidget *clist_dom;
GtkWidget *text_cmd;
GtkWidget *text_csl;

void on_clist_stk_realize (GtkWidget *widget, gpointer user_data) {
	okit_stk = OKit_StackAlloc(10) ;
	clist_stk = widget ;
}

void on_clist_stk_update (void) {
	int l = 0 ;
	int i = 0 ;
	char bnum[32] ;
	char bobj[128] ; // 128 chars should be enough
	char * bln[2] ;
	bln[0] = bnum ;
	bln[1] = bobj ;
	gtk_clist_freeze (GTK_CLIST (clist_stk)) ;
	gtk_clist_clear (GTK_CLIST (clist_stk)) ;
	l = OKit_StackDepth(okit_stk) ;
	while(i<l) {
		i ++ ;
		snprintf(bnum,30,"% 10d",i) ;
		OKit_ObjectOut(OKit_StackGet(okit_stk,i),bobj,128) ;
		gtk_clist_prepend (GTK_CLIST (clist_stk), bln) ; 
	}
	gtk_clist_thaw (GTK_CLIST (clist_stk)) ;
}

gboolean on_clist_stk_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	gint row, col ;
	gtk_clist_get_selection_info (GTK_CLIST (widget), event->x, event->y, &row, &col);
	if(event->button==2) {
//	gtk_clist_select_row (GTK_CLIST (widget), row, col) ;
		int d = OKit_StackDepth(okit_stk) ;
		if(row<d) {
			char buf[1024];
			buf[0] = 0 ;
			OKit_ObjectOut(OKit_StackGet(okit_stk,d-row),buf,1020) ;
			gtk_text_insert (GTK_TEXT (text_cmd), NULL, NULL, NULL, buf, -1);
		}
	}
	gtk_widget_grab_focus (GTK_WIDGET (text_cmd));
	return FALSE;
}


void on_clist_dom_realize (GtkWidget *widget, gpointer user_data) {
	okit_dom = OKit_NameSpaceAlloc(10) ;
	clist_dom = widget ;
}

void on_clist_dom_update (void) {
	char bnam[1024] ;  // FIXME shouldn't be a max name length
	char bobj[128] ; // 128 chars should be enough
	char * bln[2] ;
	OKit_List varl;
	OKit_Object obj;
	int varn;
	int i=0;
	bln[0] = bnam ;
	bln[1] = bobj ;
	gtk_clist_freeze (GTK_CLIST (clist_dom)) ;
	gtk_clist_clear (GTK_CLIST (clist_dom)) ;
	varl = OKit_NameSpaceOKit_List(okit_dom) ;
	varn = OKit_ListLen(varl);
	while(i<varn) {
		OKit_StrgCharString((OKit_Strg)OKit_ObjectContent(OKit_ListGet(varl,i)),bnam,1020);
		obj=OKit_NameSpaceGet(okit_dom,bnam);
		if(obj==NULL) {
			bobj[0] = 0 ;
		}
		else {
			OKit_ObjectOut(OKit_NameSpaceGet(okit_dom,bnam),bobj,128) ;
		}
		gtk_clist_prepend (GTK_CLIST (clist_dom), bln) ; 
		i ++ ;
	}
	OKit_ListFree(varl) ;
	gtk_clist_thaw (GTK_CLIST (clist_dom)) ;
}

gboolean on_clist_dom_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	gint row, col ;
	gtk_clist_get_selection_info (GTK_CLIST (widget), event->x, event->y, &row, &col);
	if(event->button==3) {
//	gtk_clist_select_row (GTK_CLIST (widget), row, col) ;
		int i = 0 ;
		char buf[1024];
		char * nam ;
		if(gtk_clist_get_cell_type(GTK_CLIST(widget),row,0)==GTK_CELL_TEXT) {
			gtk_clist_get_text(GTK_CLIST(widget),row,0,&nam);
			buf[0] = '"' ;
			while((i<1020)&&(nam[i]!=0)) {
				buf[i+1] = nam[i] ;
				i ++ ;
			}
			buf[i+1] = '"' ;
			i ++ ;
			buf[i+1] = 0 ;
			gtk_text_insert (GTK_TEXT (text_cmd), NULL, NULL, NULL, buf, -1);
		}
	}
	else if(event->button==2) {
//	gtk_clist_select_row (GTK_CLIST (widget), row, col) ;
		char buf[1024];
		char * nam ;
		if(gtk_clist_get_cell_type(GTK_CLIST(widget),row,0)==GTK_CELL_TEXT) {
			gtk_clist_get_text(GTK_CLIST(widget),row,0,&nam);
			OKit_ObjectOut(OKit_NameSpaceGet(okit_dom,nam),buf,1020) ;
			gtk_text_insert (GTK_TEXT (text_cmd), NULL, NULL, NULL, buf, -1);
		}
	}
	gtk_widget_grab_focus (GTK_WIDGET (text_cmd));
	return FALSE;
}


void on_text_cmd_realize (GtkWidget *widget, gpointer user_data) {
	text_cmd = widget;
}

int fd_error=2;

void csl_io_get (gpointer data, gint source, GdkInputCondition condition) {
	char buf[1024];
	int rd;
	while((rd=read(source,buf,1024))!=-1) {
		if(fd_error!=source) write(fd_error,buf,rd);
		gtk_text_insert (GTK_TEXT (data), NULL, NULL, NULL, buf, rd);
		if(rd<1024) break;
	}
}

void on_text_csl_realize (GtkWidget *widget, gpointer user_data) {
	int fd[2];
	text_csl = widget;
	fd_error=dup(2);
	if(fd_error==-1) {
		fd_error=2;
		fprintf(stderr,"Could not dup stderr\n");
		return;
	}
	if(pipe(fd)==-1) {
		fprintf(stderr,"Could not open console\n");
		return;
	}
	if(dup2(fd[1],2)==-1) {
		fprintf(stderr,"Could not splice console\n");
		return;
	}
	gdk_input_add(fd[0],GDK_INPUT_READ,csl_io_get,widget);
}


//################################################################################
//###
//###  Command buttons callbacks
//###



void on_button_cmd_new_clicked (GtkButton *button, gpointer user_data) {
	gtk_editable_delete_text( GTK_EDITABLE(user_data), 0, -1) ;
	gtk_widget_grab_focus (GTK_WIDGET (text_cmd));
}

void on_button_cmd_exec_clicked (GtkButton *button, gpointer user_data) {
	char * cmd;
	int pos ;
	GdkColormap *cmap;
	GdkColor *red ;
	GdkColor color ;
	cmap = gdk_colormap_get_system();
	color.red = 0xffff;
	color.green = 0;
	color.blue = 0;
	if (gdk_color_alloc(cmap, &color)) {
		red = &color ;
	}
	else {
		red = NULL ;
	}
	cmd = gtk_editable_get_chars (GTK_EDITABLE (text_cmd), 0, -1);
	pos = StrExec(cmd,opt_cur.rep_verb,okit_stk,okit_dom) ;
	on_clist_stk_update();
	on_clist_dom_update();
	gtk_editable_delete_text (GTK_EDITABLE (text_cmd), 0, -1) ;
	if(pos<0) {
		gtk_text_insert( GTK_TEXT(text_cmd),NULL,NULL,NULL,cmd,-pos) ;
		gtk_text_insert( GTK_TEXT(text_cmd),NULL,red,text_cmd->style->light,cmd-pos,-1) ;
	}
	free(cmd) ;
	gtk_widget_grab_focus (GTK_WIDGET (text_cmd));
}

gchar * keep_def_file = NULL ;

void save_data_to_file (gchar * filename, gpointer data) {
	if (filename!=NULL) {
		if (data!=NULL) {
			int fd;
			printf("Writing string to file %s\n",filename) ;
			printf("\"%s\"\n",(char *)data) ;
			fd = open (filename, O_WRONLY|O_TRUNC|O_CREAT,0644);
			if(fd==-1) {
				fprintf(stderr,"Writing to file %s FAILED\n",filename) ;
				return ;
			}
			write (fd,(char *)data,strlen((char *)data));
			write (fd,"\n",1);
			close (fd);
		}
		if(keep_def_file!=NULL) free(keep_def_file) ;
		keep_def_file = malloc(strlen(filename)+1) ;
		if(keep_def_file!=NULL) strcpy(keep_def_file,filename) ;
	}
	free(data);
}

void on_button_cmd_keep_clicked (GtkButton *button, gpointer user_data) {
	do_file_browse((keep_def_file==NULL)?"/tmp/keep.okp":keep_def_file,
			save_data_to_file,
			(gpointer)gtk_editable_get_chars (GTK_EDITABLE (text_cmd), 0, -1)) ;
	gtk_widget_grab_focus (GTK_WIDGET (text_cmd));
}



//################################################################################
//###
//###  Menu bar
//###


//###  Main menu

//void on_quit1_activate (GtkMenuItem *menuitem, gpointer user_data) {
//	gtk_widget_destroy (user_data) ;
//}

void on_options1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	opt_edt = opt_cur ;
	gtk_widget_show (dialog_options);
}


//###  File menu

gchar * file_def = NULL ;

void do_load_file (gchar * filename, gpointer data) {
	if(filename!=NULL) {
		if(file_def!=NULL) free(file_def) ;
		file_def = malloc(strlen(filename)+1) ;
		if(file_def!=NULL) strcpy(file_def,filename) ;
	}
	FileLoad(filename,opt_cur.rep_exec,okit_stk,okit_dom) ;
	on_clist_stk_update();
	on_clist_dom_update();
}

void on_load1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	do_file_browse((file_def==NULL)?"":file_def,do_load_file,NULL) ;
}


void do_exec_file (gchar * filename, gpointer data) {
	if(filename!=NULL) {
		if(file_def!=NULL) free(file_def) ;
		file_def = malloc(strlen(filename)+1) ;
		if(file_def!=NULL) strcpy(file_def,filename) ;
	}
	FileExec(filename,opt_cur.rep_exec,okit_stk,okit_dom) ;
	on_clist_stk_update();
	on_clist_dom_update();
}

void on_execute1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	do_file_browse((file_def==NULL)?"":file_def,do_exec_file,NULL) ;
}


void do_save_file (gchar * filename, gpointer data) {
	if(filename!=NULL) {
		if(file_def!=NULL) free(file_def) ;
		file_def = malloc(strlen(filename)) ;
		if(file_def!=NULL) strcpy(file_def,filename) ;
	}
	if (filename!=NULL) {
		OKit_Object obj ;
		int fd;
		obj = OKit_StackGet(okit_stk,1) ;
		printf("Writing object to file %s\n",filename) ;
		fd = open (filename, O_WRONLY|O_TRUNC|O_CREAT,0644);
		if(fd==-1) {
			fprintf(stderr,"Writing to file %s FAILED\n",filename) ;
			return ;
		}
		OKit_ObjectPrint(obj,fd) ;
		close (fd);
	}
}

void on_save1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	if(OKit_StackDepth(okit_stk)>0)
		do_file_browse((file_def==NULL)?"save.okp":file_def,do_save_file,NULL) ;
}


//###  OKit_Stack menu

void on_stk_depth1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	OKit_Object obj ;
	obj = OKit_IntgWrap(OKit_IntgNew(OKit_StackDepth(okit_stk))) ;
	OKit_StackPush(okit_stk,obj) ;
	on_clist_stk_update();
}

void on_stk_clear1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	OKit_Object obj ;
	while(OKit_StackDepth(okit_stk)) {
		obj = OKit_StackPop(okit_stk) ;
		OKit_ObjectFree(obj) ;
	}
	on_clist_stk_update();
}

OKit_Stack okit_stk_save = NULL ;

void on_stk_save1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	unsigned long i ;
	if(okit_stk_save!=NULL) {
		OKit_StackFree(okit_stk_save) ;
	}
	okit_stk_save = OKit_StackAlloc(10) ;
	i = OKit_StackDepth(okit_stk) ;
	while(i>0) {
		OKit_StackPush(okit_stk_save,OKit_StackGet(okit_stk,i)) ;
		i -- ;
	}
//	printf("FIXME: stack save not implemented\n") ;
	on_clist_stk_update();
}

void on_stk_restore1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	unsigned long i ;
	OKit_StackFree(okit_stk) ;
	okit_stk = OKit_StackAlloc(10) ;
	i = OKit_StackDepth(okit_stk_save) ;
	while(i>0) {
		OKit_StackPush(okit_stk,OKit_StackGet(okit_stk_save,i)) ;
		i -- ;
	}
//	printf("FIXME: stack restore not implemented\n") ;
	on_clist_stk_update();
}


//###  Variables menu

void on_dom_clear1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	OKit_NameSpaceFree(okit_dom) ;
	okit_dom = OKit_NameSpaceAlloc(10) ;
	on_clist_dom_update();
//	printf("FIXME: variables clear not implemented\n") ;
}

void on_freeze1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	printf("FIXME: variables freeze not implemented\n") ;
}


//###  Help menu

void on_start1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	GtkWidget *dialog_help;
	dialog_help = create_dialog_help ();
	gtk_widget_show (dialog_help);
}

void on_about1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	GtkWidget *dialog_about;
	dialog_about = create_dialog_about ();
	gtk_widget_show (dialog_about);
}

void on_syntax1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	GtkWidget *dialog_syntax;
	dialog_syntax = create_dialog_syntax ();
	gtk_widget_show (dialog_syntax);
}

void on_builtins1_activate (GtkMenuItem *menuitem, gpointer user_data) {
	GtkWidget *dialog_builtins;
	dialog_builtins = create_dialog_builtins ();
	gtk_widget_show (dialog_builtins);
}



//################################################################################
//###
//###  Toolbox
//###

void on_button_load_clicked (GtkButton *button, gpointer user_data) {
	do_file_browse((file_def==NULL)?"":file_def,do_load_file,NULL) ;
}


void on_button_exec_clicked (GtkButton *button, gpointer user_data) {
	do_file_browse((file_def==NULL)?"":file_def,do_exec_file,NULL) ;
}


void on_button_save_clicked (GtkButton *button, gpointer user_data) {
	if(OKit_StackDepth(okit_stk)>0) do_file_browse((file_def==NULL)?"save.okp":file_def,do_save_file,NULL) ;
}


void on_button_clear_clicked (GtkButton *button, gpointer user_data) {
	OKit_Object obj ;
	while(OKit_StackDepth(okit_stk)>0) {
		obj = OKit_StackPop(okit_stk) ;
		OKit_ObjectFree(obj) ;
	}
	on_clist_stk_update();
}

void on_button_undo_clicked (GtkButton *button, gpointer user_data) {
	printf("FIXME: undo not implemented\n") ;
	on_clist_stk_update();
}


void on_button_depth_clicked (GtkButton *button, gpointer user_data) {
	OKit_Object obj ;
	obj = OKit_IntgWrap(OKit_IntgNew(OKit_StackDepth(okit_stk))) ;
	OKit_StackPush(okit_stk,obj) ;
	on_clist_stk_update();
}



void on_button_dup_clicked (GtkButton *button, gpointer user_data) {
	OKit_Object obj ;
	if(OKit_StackDepth(okit_stk)>0) {
		obj = OKit_StackGet(okit_stk,1) ;
		OKit_StackPush(okit_stk,obj) ;
	}
	on_clist_stk_update();
}

void on_button_swap_clicked (GtkButton *button, gpointer user_data) {
	OKit_Object obj1 ;
	OKit_Object obj2 ;
	if(OKit_StackDepth(okit_stk)>1) {
		obj1 = OKit_StackPop(okit_stk) ;
		obj2 = OKit_StackPop(okit_stk) ;
		OKit_StackPush(okit_stk,obj1) ;
		OKit_StackPush(okit_stk,obj2) ;
	}
	on_clist_stk_update();
}

void on_button_drop_clicked (GtkButton *button, gpointer user_data) {
	OKit_Object obj ;
	if(OKit_StackDepth(okit_stk)>0) {
		obj = OKit_StackPop(okit_stk) ;
		OKit_ObjectFree(obj) ;
	}
	on_clist_stk_update();
}


//################################################################################
//###
//###  Pad buttons
//###

gboolean on_button_pad1_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "", -1);
			break ;
		case 3:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "", -1);
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad2_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "<>", -1);
			break ;
		case 3:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "<>", -1);
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad3_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "[]", -1);
			break ;
		case 3:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "[]", -1);
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad4_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "\"\"", -1);
			break ;
		case 3:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "\"\"", -1);
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad5_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "DROP", -1);
			break ;
		case 3:
			StrExec("DROP",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad6_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "DUP", -1);
			break ;
		case 3:
			StrExec("DUP",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad7_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "IFE", -1);
			break ;
		case 3:
			StrExec("IFE",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad8_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "LOOP", -1);
			break ;
		case 3:
			StrExec("LOOP",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad9_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "BRK", -1);
			break ;
		case 3:
			StrExec("BRK",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad10_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "GRP", -1);
			break ;
		case 3:
			StrExec("GRP",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad11_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "STOV", -1);
			break ;
		case 3:
			StrExec("STOV",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}

gboolean on_button_pad12_pressed (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	switch(event->button) {
		case 1:
			gtk_text_insert (GTK_TEXT (user_data), NULL, NULL, NULL, "RCLV", -1);
			break ;
		case 3:
			StrExec("RCLV",opt_cur.rep_verb,okit_stk,okit_dom) ;
			on_clist_stk_update();
			on_clist_dom_update();
			break ;
	}
	gtk_widget_grab_focus (GTK_WIDGET (user_data));
	return FALSE;
}


